/*
 * Skeleton-code behorende bij het college Netwerken, opleiding Informatica,
 * Universiteit Leiden.
 */

#include "common.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>

#include <time.h>

bool
frame_damaged(const float damaged_perc)
{
#ifdef PERFECT_CONNECTION
  /* silence compiler */
  return damaged_perc ? false : false;
#else
  float value = rand() / (float)RAND_MAX;

  return value <= (damaged_perc / 100.0) ? true : false;
#endif /* ! PERFECT_CONNECTION */
}

bool
packet_drop(const float drop_perc)
{
#ifdef PERFECT_CONNECTION
  /* silence compiler */
  return drop_perc ? false : false;
#else
  float value = rand() / (float)RAND_MAX;

  return value <= (drop_perc / 100.0) ? true : false;
#endif /* ! PERFECT_CONNECTION */
}

/* Thread-safe packet buffers shared among threads. */
struct packet_buffer_t incoming_buffer;
struct packet_buffer_t outgoing_buffer;

/* Thread that receives packets as soon as they come in. */
void *
recv_thread_func(void *ptr)
{
  int *sock_fd = (int *)ptr;

  while (true)
    {
      struct packet_t packet;

      recvfrom(*sock_fd, &packet, sizeof(struct packet_t), 0, NULL, 0);

      packet_buffer_put(&incoming_buffer, &packet, 0);
    }

  pthread_exit(0);
}

/* Thread that sends packets as soon as they enter the buffer. */
void *
send_thread_func(void *ptr)
{
  int *sock_fd = (int *)ptr;

  while (true)
    {
      struct packet_t packet;

      packet_buffer_get(&outgoing_buffer, &packet, 0);

      sendto(*sock_fd, &packet, sizeof(struct packet_t), 0, NULL, 0);
    }

  pthread_exit(0);
}

void
msleep(long msec)
{
  struct timespec t = { .tv_sec = 0, .tv_nsec = msec * 1000 * 1000};
  nanosleep(&t, NULL);
}

int
open_recv_socket(in_addr_t peer_addr, int recv_port)
{
  int sock_fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock_fd < 0)
    {
      perror("socket failed");
      exit(-1);
    }

  struct sockaddr_in input_addr = {
    .sin_addr.s_addr = peer_addr,
    .sin_family = AF_INET,
    .sin_port = htons(recv_port)
  };

  if (bind(sock_fd, (struct sockaddr *)&input_addr, sizeof(input_addr)) < 0)
    {
      perror("bind failed");
      exit(-1);
    }

  return sock_fd;
}

int
open_send_socket(in_addr_t peer_addr, int send_port)
{
  int sock_fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock_fd < 0)
    {
      perror("socket failed");
      exit(-1);
    }

  struct sockaddr_in output_addr = {
    .sin_addr.s_addr = peer_addr,
    .sin_family = AF_INET,
    .sin_port = htons(send_port)
  };

  if (connect(sock_fd, (struct sockaddr *)&output_addr, sizeof(output_addr)) < 0)
    {
      perror("connect failed");
      exit(-1);
    }

  return sock_fd;
}

bool
check_arguments(char **argv, int *port1, in_addr_t *peer_addr, int *port2)
{
  *port1 = atoi(argv[1]);
  if (*port1 <= 0 || *port1 >= 65535)
    {
      fprintf(stderr, "error: invalid port %d\n", *port1);
      return false;
    }

  *peer_addr = inet_addr(argv[2]);
  if (*peer_addr == INADDR_NONE)
    {
      fprintf(stderr, "error: invalid IP address '%s'\n", argv[2]);
      return false;
    }

  *port2 = atoi(argv[3]);
  if (*port2 <= 0 || *port2 >= 65535)
    {
      fprintf(stderr, "error: invalid port %d\n", *port2);
      return false;
    }

  return true;
}

static void
init_random_seed(void)
{
  int fh;
  unsigned seed;

  fh = open("/dev/random", O_RDONLY);
  assert(fh >= 0 && "Cannot open random device.");

  read(fh, &seed, sizeof(seed));

  close(fh);

#ifdef DEBUG
  printf("Initializing with random seed: %u\n", seed);
#endif
  srand(seed);
}

void
run(int incoming_port, in_addr_t outgoing_addr, int outgoing_port,
    MainLoopFunc main_loop)
{
  /* Initialize random seed. This can be disabled (or always intialized
   * to the same value) to get reproducible runs.
   */
  init_random_seed();

  /* Open up the network connection and launch the threads */
  int recv_sock_fd = open_recv_socket(outgoing_addr, incoming_port);
  int send_sock_fd = open_send_socket(outgoing_addr, outgoing_port);

  packet_buffer_init(&incoming_buffer);
  packet_buffer_init(&outgoing_buffer);

  pthread_t recv_thread, send_thread;
  pthread_create(&recv_thread, NULL, recv_thread_func, &recv_sock_fd);
  pthread_create(&send_thread, NULL, send_thread_func, &send_sock_fd);

  time_t start, end;

  time(&start);
  main_loop();
  time(&end);

  /* When done, forcibly shut down the threads. */
  pthread_cancel(recv_thread);
  pthread_cancel(send_thread);

  pthread_join(recv_thread, NULL);
  pthread_join(send_thread, NULL);

  close(recv_sock_fd);
  close(send_sock_fd);

  fprintf(stderr, "%ld bytes transferred per single direction in %f seconds: %f bytes/second\n",
          SR_N_PACKETS * sizeof(struct packet_t),
          difftime(end, start),
          SR_N_PACKETS * sizeof(struct packet_t) / difftime(end, start));
  fprintf(stderr, "Max. buffer utilization: send(%d), recv(%d)\n",
          outgoing_buffer.max_utilization,
          incoming_buffer.max_utilization);
}
